from __future__ import annotations

import logging
import time
from collections import Counter
from typing import List, Optional

import numpy as np
import pandas as pd

from ...constants import PROBLEM_TYPES
from ...metrics import Scorer, compute_metric, log_loss
from ...utils import get_pred_from_proba

logger = logging.getLogger(__name__)


class AbstractWeightedEnsemble:
    def predict(self, X):
        y_pred_proba = self.predict_proba(X)
        return get_pred_from_proba(y_pred_proba=y_pred_proba, problem_type=self.problem_type)

    def predict_proba(self, X):
        return self.weight_pred_probas(X, weights=self.weights_)

    @staticmethod
    def weight_pred_probas(pred_probas, weights):
        preds_norm = [pred * weight for pred, weight in zip(pred_probas, weights) if weight != 0]
        preds_ensemble = np.sum(preds_norm, axis=0)
        return preds_ensemble


class EnsembleSelection(AbstractWeightedEnsemble):
    def __init__(
        self,
        ensemble_size: int,
        problem_type: str,
        metric: Scorer,
        sorted_initialization: bool = False,
        bagging: bool = False,
        tie_breaker: str = "random",
        subsample_size: int | None = None,
        random_state: Optional[np.random.RandomState] = None,
        **kwargs,
    ):
        self.ensemble_size = ensemble_size
        self.problem_type = problem_type
        self.metric = metric
        self.sorted_initialization = sorted_initialization
        self.bagging = bagging
        self.use_best = True
        if tie_breaker not in ["random", "second_metric"]:
            raise ValueError(f"Unknown tie_breaker value: {tie_breaker}. Must be one of: ['random', 'second_metric']")
        self.tie_breaker = tie_breaker
        self.subsample_size = subsample_size
        if random_state is not None:
            self.random_state = random_state
        else:
            self.random_state = np.random.RandomState(seed=0)
        self.quantile_levels = kwargs.get("quantile_levels", None)

    def fit(
        self, predictions: List[np.ndarray], labels: np.ndarray, time_limit=None, identifiers=None, sample_weight=None
    ):
        self.ensemble_size = int(self.ensemble_size)
        if self.ensemble_size < 1:
            raise ValueError("Ensemble size cannot be less than one!")
        if not self.problem_type in PROBLEM_TYPES:
            raise ValueError("Unknown problem type %s." % self.problem_type)
        # if not isinstance(self.metric, Scorer):
        #     raise ValueError('Metric must be of type scorer')

        self._fit(predictions=predictions, labels=labels, time_limit=time_limit, sample_weight=sample_weight)
        self._calculate_weights()
        logger.log(15, "Ensemble weights: ")
        logger.log(15, self.weights_)
        return self

    # TODO: Consider having a removal stage, remove each model and see if score is affected, if improves or not effected, remove it.
    def _fit(self, predictions: List[np.ndarray], labels: np.ndarray, time_limit=None, sample_weight=None):
        ensemble_size = self.ensemble_size
        if isinstance(labels, pd.Series):
            labels = labels.values
        self.num_input_models_ = len(predictions)
        ensemble = []
        trajectory = []
        order = []
        used_models = set()
        num_samples_total = len(labels)

        if self.subsample_size is not None and self.subsample_size < num_samples_total:
            logger.log(15, f"Subsampling to {self.subsample_size} samples to speedup ensemble selection...")
            subsample_indices = self.random_state.choice(num_samples_total, self.subsample_size, replace=False)
            labels = labels[subsample_indices]
            for i in range(self.num_input_models_):
                predictions[i] = predictions[i][subsample_indices]

        # if self.sorted_initialization:
        #     n_best = 20
        #     indices = self._sorted_initialization(predictions, labels, n_best)
        #     for idx in indices:
        #         ensemble.append(predictions[idx])
        #         order.append(idx)
        #         ensemble_ = np.array(ensemble).mean(axis=0)
        #         ensemble_performance = calculate_score(
        #             labels, ensemble_, self.task_type, self.metric,
        #             ensemble_.shape[1])
        #         trajectory.append(ensemble_performance)
        #     ensemble_size -= n_best

        time_start = time.time()
        round_scores = False
        epsilon = 1e-4
        round_decimals = 6
        ensemble_prediction = np.zeros(predictions[0].shape)
        weighted_ensemble_prediction = np.zeros(predictions[0].shape)
        fant_ensemble_prediction = np.zeros(predictions[0].shape)
        for i in range(ensemble_size):
            scores = np.zeros((len(predictions)))
            s = len(ensemble)

            if s != 0:
                ensemble_prediction *= (s - 1) / s
                ensemble_prediction += ensemble[-1] / s
                weighted_ensemble_prediction[:] = (s / float(s + 1)) * ensemble_prediction
            for j, pred in enumerate(predictions):
                fant_ensemble_prediction[:] = weighted_ensemble_prediction + (1.0 / float(s + 1)) * pred
                if self.problem_type in ["multiclass", "softclass"]:
                    # Renormalize
                    fant_ensemble_prediction /= fant_ensemble_prediction.sum(axis=1)[:, np.newaxis]
                scores[j] = self._calculate_regret(
                    y_true=labels,
                    y_pred_proba=fant_ensemble_prediction,
                    metric=self.metric,
                    sample_weight=sample_weight,
                )
                if round_scores:
                    scores[j] = scores[j].round(round_decimals)

            all_best = np.argwhere(np.isclose(scores, np.nanmin(scores), atol=0, rtol=1e-12)).flatten()

            if (len(all_best) > 1) and used_models:
                # If tie, prioritize models already in ensemble to avoid unnecessarily large ensemble
                new_all_best = []
                for m in all_best:
                    if m in used_models:
                        new_all_best.append(m)
                if new_all_best:
                    all_best = new_all_best

            if len(all_best) > 1:
                if self.tie_breaker == "second_metric":
                    if self.problem_type in ["binary", "multiclass"]:
                        # Tiebreak with log_loss
                        scores_tiebreak = np.zeros((len(all_best)))
                        secondary_metric = log_loss
                        fant_ensemble_prediction = np.zeros(weighted_ensemble_prediction.shape)
                        index_map = {}
                        for k, j in enumerate(all_best):
                            index_map[k] = j
                            pred = predictions[j]
                            fant_ensemble_prediction[:] = weighted_ensemble_prediction + (1.0 / float(s + 1)) * pred
                            scores_tiebreak[k] = self._calculate_regret(
                                y_true=labels, y_pred_proba=fant_ensemble_prediction, metric=secondary_metric
                            )
                        all_best_tiebreak = np.argwhere(
                            np.isclose(scores_tiebreak, np.nanmin(scores_tiebreak), atol=0, rtol=1e-12)
                        ).flatten()
                        all_best = [index_map[index] for index in all_best_tiebreak]

            best = self.random_state.choice(all_best)
            best_score = scores[best]

            # If first iteration
            if i == 0:
                # If abs value of min score is large enough, round to 6 decimal places to avoid floating point error deciding the best index.
                # This avoids 2 models with the same pred proba both being used in the ensemble due to floating point error
                if np.abs(best_score) > epsilon:
                    round_scores = True
                    best_score = best_score.round(round_decimals)

            ensemble.append(predictions[best])
            trajectory.append(best_score)
            order.append(best)
            used_models.add(best)

            # Handle special case
            if len(predictions) == 1:
                break

            if time_limit is not None:
                time_elapsed = time.time() - time_start
                time_left = time_limit - time_elapsed
                if time_left <= 0:
                    logger.warning(
                        "Warning: Ensemble Selection ran out of time, early stopping at iteration %s. This may mean that the time_limit specified is very small for this problem."
                        % (i + 1)
                    )
                    break

        min_score = np.min(trajectory)
        first_index_of_best = trajectory.index(min_score)

        if self.use_best:
            self.indices_ = order[: first_index_of_best + 1]
            self.trajectory_ = trajectory[: first_index_of_best + 1]
            self.train_score_ = trajectory[first_index_of_best]
            self.ensemble_size = first_index_of_best + 1
            logger.log(15, "Ensemble size: %s" % self.ensemble_size)
        else:
            self.indices_ = order
            self.trajectory_ = trajectory
            self.train_score_ = trajectory[-1]

        logger.debug("Ensemble indices: " + str(self.indices_))

    def _calculate_regret(
        self, y_true: np.ndarray, y_pred_proba: np.ndarray, metric: Scorer, sample_weight: np.ndarray = None
    ) -> float:
        if metric.needs_pred or metric.needs_quantile:
            y_pred = get_pred_from_proba(y_pred_proba=y_pred_proba, problem_type=self.problem_type)
            y_pred_proba = None
        else:
            y_pred = None
        regret = compute_metric(
            y=y_true,
            y_pred=y_pred,
            y_pred_proba=y_pred_proba,
            metric=metric,
            weights=sample_weight,
            quantile_levels=self.quantile_levels,
            as_error=True,
        )
        return regret

    def _calculate_weights(self):
        ensemble_members = Counter(self.indices_).most_common()
        weights = np.zeros((self.num_input_models_,), dtype=float)
        for ensemble_member in ensemble_members:
            weight = float(ensemble_member[1]) / self.ensemble_size
            weights[ensemble_member[0]] = weight

        if np.sum(weights) < 1:
            weights = weights / np.sum(weights)

        self.weights_ = weights


class SimpleWeightedEnsemble(AbstractWeightedEnsemble):
    """Predefined user-weights ensemble"""

    def __init__(self, weights, problem_type, **kwargs):
        self.weights_ = weights
        self.problem_type = problem_type

    @property
    def ensemble_size(self):
        return len(self.weights_)



from __future__ import annotations

import copy
import logging
import math
import time
from dataclasses import dataclass
from typing import Callable, Literal

import pandas as pd

from autogluon.common.utils.resource_utils import get_resource_manager
from autogluon.core.models import AbstractModel, BaggedEnsembleModel

logger = logging.getLogger(__name__)

DEFAULT_REMOTE_KWARGS = dict(max_calls=1, retry_exceptions=False, max_retries=0)


@dataclass
class ModelResources:
    """Resource allocation for a model."""

    num_gpus_for_fold_worker: int
    num_cpus_for_fold_worker: int
    num_gpus_for_model_worker: int
    num_cpus_for_model_worker: int
    total_num_cpus: int
    total_num_gpus: int


class ParallelFitManager:
    """Tracks how many resources are used when scheduling jobs in a parallel setting.

    Parallel Fit
    ---------------
    We use ray to start a model-worker with a certain number of CPUs and GPUs.
    The worker then starts new fold-workers to fit each fold of a model.
    Alternatively, the model-worker uses the given resources to perform a single fit.

    We must pass GPU resources to a model-worker if the model has `refit_folds is True`
    as the refit_folds call happens in the model-worker.

    For full parallelization, we require the following:
        - GPUs
            - refit_folds is True: `num_gpus` + `num_bag_folds` * `num_bag_sets` * `num_gpus`
            - refit_folds is False: `num_bag_folds` * `num_bag_sets` * `num_gpus`
        - CPUs:
            - model with bagging: 1 + `num_cpus` * `num_bag_folds` * `num_bag_sets`
            - model without bagging: `num_cpus`

    Parameters
    ----------
    mode: {"fit", "refit"}
        The mode to use for fitting the models.
    func: callable
        The fit function to distribute.
    func_kwargs: dict, default=None
        Additional kwargs to pass to the function.
    func_put_kwargs: dict, default=None
        Additional kwargs to pass to the function, where the values are put into the object store.
    num_cpus : int | str
        Total number of CPUs available in the cluster (or `auto`).
    num_gpus : int | str
        Total number of GPUs available in the cluster (or `auto`).
    num_splits : int | None, default=None
        Number of training splits/bags for a model. Required if mode='fit'.
    get_model_attribute_func : callable, default=None
        Function to get an attribute for a model. Required if mode='refit'.
    """

    def __init__(
        self,
        *,
        mode: Literal["fit", "refit"],
        func: Callable,
        func_kwargs: dict,
        func_put_kwargs: dict,
        num_cpus: int | str,
        num_gpus: int | str,
        num_splits: int | None = None,
        get_model_attribute_func: Callable | None = None,
        X: pd.DataFrame,
        y: pd.Series,
        problem_type="infer",
        num_classes="infer",
        total_mem: int | None | str = "auto",
        max_mem_frac: float = 0.8,
        delay_between_jobs: float = 0,
    ):
        self.X = X
        self.y = y
        self.problem_type = problem_type
        self.num_classes = num_classes
        self.delay_between_jobs = delay_between_jobs
        import ray

        self.mode = mode
        self.num_splits = num_splits
        self.get_model_attribute_func = get_model_attribute_func

        if self.mode == "fit":
            assert num_splits is not None, "num_splits must be set for mode='fit'."
        if self.mode == "refit":
            assert get_model_attribute_func is not None, "get_model_attribute_func must be set for mode='refit'."

        # Resource tracking
        if isinstance(num_cpus, str):
            num_cpus = get_resource_manager().get_cpu_count()
        if isinstance(num_gpus, str):
            num_gpus = get_resource_manager().get_gpu_count()
        self.total_num_cpus = num_cpus
        self.total_num_gpus = num_gpus
        self.available_num_cpus = num_cpus
        self.available_num_gpus = num_gpus
        # self.extra_num_cpus = self.total_num_cpus // 10  # FIXME: Maybe remove, allows for oversubscribing
        self.extra_num_cpus = 0  # Setting to 0 for now to avoid using more resources than requested by the user
        if isinstance(total_mem, str) and total_mem == "auto":
            # FIXME: when None should be infinite, implement
            total_mem = get_resource_manager().get_available_virtual_mem()
        # FIXME: Should this be passed during init or re-evaluated for each schedule call?
        self.total_mem = total_mem
        self.available_mem = self.total_mem
        self.max_mem_frac = max_mem_frac
        self.max_mem = self.total_mem * self.max_mem_frac

        # Detect max resources a job could use on some node
        self.max_cpu_resources_per_node = int(max([n["Resources"]["CPU"] for n in ray.state.nodes()]))

        # Job tracking
        self.job_refs_to_allocated_resources: dict[str, ModelResources] = {}
        self.job_refs_to_model_name: dict[str, str] = {}
        self.job_refs_to_model_memory_estimate: dict[str, int] = {}
        self.model_child_mem_estimate_cache: dict[str, int] = {}
        self.models_to_schedule: list[AbstractModel] | list[str] = []

        # Init remote function
        self.remote_func = ray.remote(**DEFAULT_REMOTE_KWARGS)(func)
        self.job_kwargs = dict()
        for key, value in func_kwargs.items():
            self.job_kwargs[key] = value
        for key, value in func_put_kwargs.items():
            self.job_kwargs[key] = ray.put(value)
        self.func_put_kwargs = func_put_kwargs

        logger.log(
            20,
            f"ParallelFitManager Details:"
            f"\n\tCPU Total: {self.total_num_cpus}"
            f"\n\tGPU Total: {self.total_num_gpus}"
            f"\n\tMem Total: {self.total_mem * 1e-9:.1f} GB"
            f"\n\t    Max Allowed: {self.max_mem * 1e-9:.1f}/{self.total_mem * 1e-9:.1f} GB (max_mem_frac={self.max_mem_frac})"
            f"\n\t    Max Allowed Per Core: {self.max_mem_per_core * 1e-9:.2f}/{self.total_mem_per_core * 1e-9:.2f} GB",
        )

    @property
    def available_num_cpus_virtual(self) -> int:
        return self.available_num_cpus + self.extra_num_cpus

    @property
    def total_num_cpus_virtual(self) -> int:
        return self.total_num_cpus + self.extra_num_cpus

    def num_children_model(self, model: AbstractModel) -> int:
        if (not isinstance(model, BaggedEnsembleModel)) or model._user_params.get("use_child_oof", False):
            return 1
        else:
            return self.num_splits

    @property
    def max_mem_per_core(self):
        return self.max_mem / self.total_num_cpus

    @property
    def total_mem_per_core(self):
        return self.total_mem / self.total_num_cpus

    # FIXME: Don't mutate models ever
    #  Mutate num_cpus user args in worker process or even better just pass num_cpus as fit kwarg
    # TODO: Don't give LightGBM/XGBoost/CatBoost more CPUs than they want, cap them at 50% of resources no matter what
    # TODO: We can lazily execute this logic. We first come up with a plan, then we schedule the workers. This allows for optimal CPU utilization.
    # FIXME v1.2: Use available memory to re-calculate per-core values, will more effectively use memory.
    # TODO: Test performance if not allowing overallocated CPUs
    def schedule_jobs(self, *, models_to_fit: list[AbstractModel] | list[str] | None = None) -> list[str]:
        """Schedule model training.

        This function must be first called with `models_to_fit is not None` and then with `models_to_fit is None`.
        Whereby the first call initializes the list of models to fit and subsequent calls schedule the remaining jobs.

        models_to_fit: list[AbstractModel] | list[str] | None, default=None
            The models that shall be fitted in a distributed manner.
        """
        import ray

        max_mem_per_core = self.max_mem_per_core

        if models_to_fit is not None:
            models_to_schedule = models_to_fit
        else:
            models_to_schedule = self.models_to_schedule

        cpus_fully_allocated = False
        num_models_delay_to_fit_all = 0

        job_refs = []
        num_models_to_schedule = len(models_to_schedule)
        models_to_schedule_tmp = []
        total_models_to_fit = 0
        for i, model in enumerate(models_to_schedule):
            model_name = model if self.mode == "refit" else model.name
            if model_name in self.model_child_mem_estimate_cache:
                model_child_memory_estimate = self.model_child_mem_estimate_cache[model_name]
            else:
                try:
                    # FIXME: DONT USE TRY/EXCEPT, this is done to handle models crashing during initialization such as KNN when `NoValidFeatures`. Instead figure this out earlier or in the worker thread
                    model_child_memory_estimate = self.get_memory_estimate_for_model_child(model=model)
                except Exception as e:
                    logger.log(
                        20,
                        f"Ran into exception when getting memory estimate for model, skipping model {model.name}: {e.__class__.__name__}: {e}",
                    )
                    continue
                self.model_child_mem_estimate_cache[model_name] = model_child_memory_estimate
            if model_child_memory_estimate > self.max_mem:
                logger.log(
                    20, f"Insufficient total memory to fit model for even a single fold. Skipping {model_name}..."
                )
                continue

            num_children = self.num_children_model(model=model)
            total_models_to_fit += num_children
            models_to_schedule_tmp.append(model)

        if total_models_to_fit == 0:
            self.models_to_schedule = []
            return []

        # Use real CPUs to avoid overallocating
        available_cpus = self.available_num_cpus
        cpus_per_model = available_cpus / total_models_to_fit
        cpus_per_model = math.floor(cpus_per_model)

        # We should always use at least this many CPUs per model
        # TODO: This isn't optimal if the user provides specific CPU requirements per model, we should take this into account
        num_cpus_per_child_floor = max(int(cpus_per_model), 1)

        models_to_schedule = models_to_schedule_tmp
        models_to_schedule_later = []
        for i, model in enumerate(models_to_schedule):
            model_name = model if self.mode == "refit" else model.name
            # TODO: refactor memory estimate logic to be one function call with the same code below
            if model_name in self.model_child_mem_estimate_cache:
                model_child_memory_estimate = self.model_child_mem_estimate_cache[model_name]
            else:
                try:
                    # FIXME: DONT USE TRY/EXCEPT, this is done to handle models crashing during initialization such as KNN when `NoValidFeatures`. Instead figure this out earlier or in the worker thread
                    model_child_memory_estimate = self.get_memory_estimate_for_model_child(model=model)
                except Exception as e:
                    logger.log(
                        20,
                        f"Ran into exception when getting memory estimate for model, skipping model {model.name}: {e.__class__.__name__}: {e}",
                    )
                    continue
                self.model_child_mem_estimate_cache[model_name] = model_child_memory_estimate
            if model_child_memory_estimate > self.max_mem:
                logger.log(
                    20, f"Insufficient total memory to fit model for even a single fold. Skipping {model_name}..."
                )
                continue
            if self.available_num_cpus_virtual < 1:
                if not cpus_fully_allocated:
                    logger.log(15, f"Delay scheduling {num_models_to_schedule - i} models: CPUs are fully allocated")
                cpus_fully_allocated = True
                models_to_schedule_later.append(model)
                continue
            num_children = self.num_children_model(model=model)
            if (num_children > self.available_num_cpus_virtual) and (self.total_num_cpus >= num_children):
                # try to wait to schedule later when all folds can be fit in parallel
                num_models_delay_to_fit_all += 1
                models_to_schedule_later.append(model)
                continue
            if num_models_delay_to_fit_all > 0:
                logger.log(
                    15,
                    f"Delay scheduling {num_models_delay_to_fit_all} models: waiting for enough CPUs to fit all folds in parallel...",
                )
                num_models_delay_to_fit_all = 0

            num_cpus_per_child_safe = model_child_memory_estimate / max_mem_per_core
            num_cpus_per_child_safe = max(math.ceil(num_cpus_per_child_safe), num_cpus_per_child_floor)

            num_cpus_avail = self.available_num_cpus_virtual

            # FIXME: Adjust num_cpus_per_child if fitting fewer folds in parallel?
            # FIXME: Not accurate, due to oversubscription, this is dangerous for memory...
            max_safe_children = math.floor(num_cpus_avail / num_cpus_per_child_safe)

            safe_children = max(min(max_safe_children, num_children), 0)

            if safe_children < num_children:
                # FIXME: Make this better, do real successive halving rather than this hack code that only works for 8 or fewer
                if safe_children >= 8:
                    safe_children = 8
                elif safe_children >= 4:
                    safe_children = 4
                elif safe_children >= 2:
                    safe_children = 2
                elif safe_children >= 1:
                    safe_children = 1
                else:
                    safe_children = 0
                    # skip model
                    pass

            if safe_children == 0:
                logger.log(15, f"Delay scheduling model {model_name}: No safe children able to be fit.")
                models_to_schedule_later.append(model)
                continue

            model_memory_estimate = self.get_memory_estimate_for_model(
                model=model, mem_usage_child=model_child_memory_estimate, num_children=safe_children
            )

            if safe_children < num_children:
                if ((num_children * model_child_memory_estimate) < self.max_mem) and (
                    self.total_num_cpus >= num_children
                ):
                    # try to wait to schedule later when all folds can be fit in parallel
                    logger.log(
                        15,
                        f"Delay scheduling model {model_name}: Currently can safely fit {safe_children} folds in parallel, "
                        f"waiting to be able to fit all {num_children} folds in parallel.",
                    )
                    models_to_schedule_later.append(model)
                    continue
                else:
                    logger.log(
                        20,
                        f"NOTE: {model_name} is too large to ever fit all {num_children} folds in parallel. Fitting {safe_children} folds in parallel...",
                    )
                    # Will never be able to fit all children in parallel because it would use too much memory
                    # TODO: Figure out best option here, for now we train them immediately
                    #  but it may be better to deprioritize them in favor of lower memory usage models

            if isinstance(model, AbstractModel):
                # FIXME: This does in-place mutation of model, don't do this!
                # FIXME: Prefer to avoid this hack, but doing so would require some refactoring.
                #  Basically, we want to avoid passing `model` when it is initialized.
                #  Rather, it is better to pass the cls and kwargs to initialize the model downstream as late as possible.
                model = prepare_model_resources_for_fit(
                    model=model,
                    num_cpus=num_cpus_per_child_safe,
                    num_gpus=0,
                    num_parallel=safe_children,
                    num_children=num_children,
                    total_num_cpus=self.total_num_cpus,
                    total_num_gpus=self.total_num_gpus,
                )
            model_resources = self.get_resources_for_model(model=model)

            # FIXME: Keep track of memory usage estimates, use it to limit # of parallel model fits
            # FIXME: Should this logic ever trigger given the above new logic that checks for `safe_children`
            is_sufficient, reason = self.check_sufficient_resources(resources=model_resources)
            if not is_sufficient:
                if len(job_refs) + len(self.job_refs_to_allocated_resources) == 0:
                    logger.log(
                        20,
                        "DISTRIBUTED WARNING: Insufficient total resources for training a model fully distributed parallel. "
                        "Consider disabling distributed training. "
                        "Forcing to train one model anyhow, but this will lead to inefficient parallelization.",
                    )

                    # Ray's nested calls will keep blocking GPUs and thus create a deadlock if all GPUs are allocated to the model-worker and
                    # none can be used by the fold-worker.
                    if (
                        model_resources.num_gpus_for_model_worker + model_resources.num_gpus_for_fold_worker
                    ) > self.total_num_gpus:
                        raise ValueError(
                            "DISTRIBUTED ERROR: Insufficient number of GPUs to train any model, even in a non-parallel setting. "
                            "This is likely the results of requiring more GPUs than available to distribute the training. "
                            "Ray does not support freeing GPU resources for nested calls with GPUs. "
                            "Thus, we need at least twice the amount of GPUs needed to fit one model."
                        )
                else:
                    if (
                        model_resources.num_gpus_for_model_worker + model_resources.num_gpus_for_fold_worker
                    ) > self.total_num_gpus:
                        logger.log(
                            40,
                            f"DISTRIBUTED WARNING: Delay scheduling model {model_name}: "
                            "Insufficient number of GPUs to train any model, even in a non-parallel setting. "
                            "This is likely the results of requiring more GPUs than available to distribute the training. "
                            "Ray does not support freeing GPU resources for nested calls with GPUs. "
                            "Thus, we need at least twice the amount of GPUs needed to fit one model.",
                        )

                    logger.log(15, f"Delay scheduling model {model_name}: {reason}.")
                    models_to_schedule_later.append(model)
                    continue

            job_ref = self.remote_func.options(
                num_cpus=model_resources.num_cpus_for_model_worker, num_gpus=model_resources.num_gpus_for_model_worker
            ).remote(model=ray.put(model) if self.mode in ["fit"] else model, **self.job_kwargs)
            job_refs.append(job_ref)
            self.allocate_resources(
                job_ref=job_ref,
                resources=model_resources,
                model_name=model_name,
                model_memory_estimate=model_memory_estimate,
            )

            logger.log(
                20,
                f"Scheduled {model_name}: "
                f"allocated {'' if is_sufficient else 'UP TO '}{model_resources.total_num_cpus} CPUs and {model_resources.total_num_gpus} GPUs | "
                f"{len(self.job_refs_to_allocated_resources)} jobs running"
                f"\n\t{model_resources.num_cpus_for_fold_worker if num_children != 1 else model_resources.num_cpus_for_model_worker} CPUs each for {num_children} folds, fitting {safe_children} in parallel"
                f"\n\t{self.total_num_cpus - self.available_num_cpus}/{self.total_num_cpus} Allocated CPUS"
                f"\t| {(self.total_mem - self.available_mem) * 1e-9:.1f}/{self.total_mem * 1e-9:.1f} GB Allocated Memory",
            )
            if self.delay_between_jobs > 0:
                time.sleep(self.delay_between_jobs)

        if num_models_delay_to_fit_all > 0:
            logger.log(
                15,
                f"Delay scheduling {num_models_delay_to_fit_all} models: waiting for enough CPUs to fit all folds in parallel...",
            )
            num_models_delay_to_fit_all = 0

        self.models_to_schedule = models_to_schedule_later
        return job_refs

    def check_sufficient_resources(self, *, resources: ModelResources) -> tuple[bool, str | None]:
        """Determine if there are enough resources to scheduling fitting another model."""

        # Allow for oversubscribing to 10% of the CPUs due to scheduling overhead.
        if self.available_num_cpus_virtual < resources.total_num_cpus:
            return False, "not enough CPUs free."

        # All models need at least one CPU but not all a GPU
        # Avoid scheduling a model if there are not at least 50% of the required GPUs available
        if (resources.total_num_gpus > 0) and (self.available_num_gpus < math.ceil(resources.total_num_gpus / 2)):
            return False, "not enough GPUs free."

        return True, None

    def get_resources_for_model(self, *, model: AbstractModel | str) -> ModelResources:
        if self.mode == "fit":
            # model is AbstractModel
            return self.get_resources_for_model_fit(model=model)
        elif self.mode == "refit":
            # model is str
            return self.get_resources_for_model_refit(model=model)
        else:
            raise ValueError(f"Unknown mode: {self.mode}")

    def get_memory_estimate_for_model_child(self, *, model: AbstractModel) -> int:
        X = self.X
        y = self.y

        if model.can_estimate_memory_usage_static_child():
            # This logic is fast and optimized
            # FIXME: Don't use `_user_params`
            if isinstance(model, BaggedEnsembleModel):
                hyperparameters = model.model_base._user_params
            else:
                hyperparameters = model._user_params
            mem_usage_child = model.estimate_memory_usage_static_child(
                X=X,
                y=y,
                hyperparameters=hyperparameters,
                problem_type=self.problem_type,
                num_classes=self.num_classes,
            )
            return mem_usage_child
        else:
            # This logic is over 100x slower than the static method. Not recommended.
            model_clone = copy.deepcopy(model)
            model_clone.initialize(
                X=X,
                y=y,
                problem_type=self.problem_type,
                num_classes=self.num_classes,
            )
            if isinstance(model, BaggedEnsembleModel):
                model_clone.model_base.initialize(
                    X=X,
                    y=y,
                    problem_type=self.problem_type,
                    num_classes=self.num_classes,
                )
            mem_usage_child = model_clone.estimate_memory_usage_child(
                X=X,
                y=y,
            )

            return mem_usage_child

    def get_memory_estimate_for_model(
        self, *, model: AbstractModel, mem_usage_child: int = None, num_children: int = None
    ) -> int:
        if num_children is None:
            num_children = self.num_children_model(model)
        if mem_usage_child is None:
            mem_usage_child = self.get_memory_estimate_for_model_child(model=model)
        mem_usage_bag = mem_usage_child * num_children
        mem_usage_child_mb = mem_usage_child * 1e-6
        mem_usage_bag_mb = mem_usage_child_mb * num_children

        logger.log(
            15,
            f"\t{mem_usage_bag_mb:.0f} MB (per bag)\t| {mem_usage_child_mb:.0f} MB (per child)\t| {num_children} children\t| {model.name}",
        )
        return mem_usage_bag

    def get_resources_for_model_refit(self, model: str) -> ModelResources:
        """Estimate the resources required to fit a model."""

        # FIXME: We should synchronize this with the actual values used downstream.
        #  Currently we are specifying it here and downstream separately, which gives the opportunity for them to not be matching.
        #  This should be fixed.
        # FIXME: Should this use fit_num_cpus_child or fit_num_cpus ?
        num_gpus_for_fold_worker = self.get_model_attribute_func(model=model, attribute="fit_num_gpus_child")
        num_cpus_for_fold_worker = self.get_model_attribute_func(model=model, attribute="fit_num_cpus_child")
        num_cpus_for_fold_worker = (
            num_cpus_for_fold_worker
            if num_cpus_for_fold_worker is not None
            else min(self.max_cpu_resources_per_node, self.total_num_cpus)
        )
        num_gpus_for_fold_worker = num_gpus_for_fold_worker if num_gpus_for_fold_worker is not None else 0

        num_gpus_for_model_worker = (
            1 if self.get_model_attribute_func(model=model, attribute="refit_full_requires_gpu") else 0
        )
        return ModelResources(
            num_gpus_for_fold_worker=num_gpus_for_fold_worker,
            num_cpus_for_fold_worker=num_cpus_for_fold_worker,
            num_gpus_for_model_worker=num_gpus_for_model_worker,
            num_cpus_for_model_worker=1,
            # num_cpus_for_model_worker is freed once the nested ray call is done
            total_num_cpus=num_cpus_for_fold_worker,
            total_num_gpus=num_gpus_for_model_worker + num_gpus_for_fold_worker,
        )

    # FIXME: Refactor this to be cleaner after initial merge. This code works but it is overly confusing.
    def get_resources_for_model_fit(self, *, model: AbstractModel) -> ModelResources:
        """Estimate the resources required to fit a model."""

        if "num_cpus" not in getattr(model, "model_base", model)._user_params_aux:
            logger.warning(
                f"DISTRIBUTED WARNING: Model {model.name} does not specify the number of resources to use! "
                "Assuming that the model will use all available node resources, which can heavily impact the performance of distributed training."
            )

        # Fallback if required information are not given at this point, we can only assume that the model will use all available CPU resources of a node.
        num_cpus_for_fold_worker = getattr(model, "model_base", model)._user_params_aux.get(
            "num_cpus", self.max_cpu_resources_per_node
        )
        # As we have no information about if the model needs GPUs, we assume that it does not need any.
        num_gpus_for_fold_worker = getattr(model, "model_base", model)._user_params_aux.get("num_gpus", 0)

        if (not isinstance(model, BaggedEnsembleModel)) or model._user_params.get("use_child_oof", False):
            # Only one fold is fit, so we need to use all fit resources for the model-worker
            num_cpus_for_model_worker = num_cpus_for_fold_worker
            num_gpus_for_model_worker = num_gpus_for_fold_worker
            num_cpus_for_fold_worker = 0
            num_gpus_for_fold_worker = 0
            total_num_cpus = num_cpus_for_model_worker
        else:
            # If refit_folds is True, we need to pass GPU resources to the model-worker
            num_gpus_for_model_worker = (
                num_gpus_for_fold_worker
                if ((num_gpus_for_fold_worker > 0) and model._user_params.get("refit_folds", False))
                else 0
            )
            num_cpus_for_model_worker = 1

            # num_cpus_for_model_worker is freed once the nested ray call is done
            total_num_cpus = model._user_params_aux["num_cpus"]

        return ModelResources(
            num_gpus_for_fold_worker=num_gpus_for_fold_worker,
            num_cpus_for_fold_worker=num_cpus_for_fold_worker,
            num_gpus_for_model_worker=num_gpus_for_model_worker,
            num_cpus_for_model_worker=num_cpus_for_model_worker,
            total_num_cpus=total_num_cpus,
            total_num_gpus=num_gpus_for_model_worker + num_gpus_for_fold_worker * self.num_splits,
        )

    def allocate_resources(
        self, *, job_ref: str, resources: ModelResources, model_memory_estimate: int, model_name: str = None
    ) -> None:
        """Allocate resources for a model fit."""

        self.available_num_cpus -= resources.total_num_cpus
        self.available_num_gpus -= resources.total_num_gpus
        self.available_mem -= model_memory_estimate
        self.job_refs_to_allocated_resources[job_ref] = resources
        self.job_refs_to_model_name[job_ref] = model_name
        self.job_refs_to_model_memory_estimate[job_ref] = model_memory_estimate

    def deallocate_resources(self, *, job_ref: str) -> None:
        """Deallocate resources for a model fit."""

        resources = self.job_refs_to_allocated_resources.pop(job_ref)
        model_name = self.job_refs_to_model_name.pop(job_ref)
        self.available_num_cpus += resources.total_num_cpus
        self.available_num_gpus += resources.total_num_gpus
        model_memory_estimate = self.job_refs_to_model_memory_estimate.pop(job_ref)
        self.available_mem += model_memory_estimate
        self.model_child_mem_estimate_cache.pop(model_name)

    def clean_unfinished_job_refs(self, *, unfinished_job_refs: list[str] | None = None):
        import ray

        # TODO: determine how to suppress error messages from cancelling jobs.
        if unfinished_job_refs is not None and len(unfinished_job_refs) > 0:
            model_names = [self.job_refs_to_model_name[f] for f in unfinished_job_refs]
            logger.log(20, f"Cancelling {len(model_names)} jobs: {model_names}")
            for f in unfinished_job_refs:
                ray.cancel(f, force=True)

    def clean_job_state(self, *, unfinished_job_refs: list[str] | None = None) -> None:
        """Clean up state of manager."""
        self.job_refs_to_allocated_resources = {}
        self.job_refs_to_model_name = {}
        self.job_refs_to_model_memory_estimate = {}
        self.model_child_mem_estimate_cache = {}
        self.models_to_schedule = []
        self.available_num_cpus = self.total_num_cpus
        self.available_num_gpus = self.total_num_gpus
        self.available_mem = self.total_mem
        self.clean_unfinished_job_refs(unfinished_job_refs=unfinished_job_refs)

    def clean_up_ray(self, *, unfinished_job_refs: list[str] | None = None) -> None:
        """Try to clean up ray object store."""
        import ray

        self.clean_unfinished_job_refs(unfinished_job_refs=unfinished_job_refs)

        ray.internal.free(object_refs=[self.job_kwargs[key] for key in self.func_put_kwargs])
        for key in self.func_put_kwargs:
            del self.job_kwargs[key]


# TODO: This logic is a hack. We shouldn't be editing models in-place. Refactor this as a fast-follow. Currently it works, but it is not ideal long term.
def prepare_model_resources_for_fit(
    *,
    model: AbstractModel,
    total_num_cpus: int,
    total_num_gpus: int,
    num_cpus: int = 1,
    num_gpus: float = 0,
    num_cpus_worker: int = 1,
    num_gpus_worker: float = 0,
    num_parallel: int = None,
    num_children: int = 1,
) -> AbstractModel:
    """Allocate each model resources for fitting. (This is currently an in-place operation!)

    We allocate resources by setting the _user_params_aux of a model.

    """
    if num_parallel is None:
        num_parallel = num_children
    assert num_parallel <= num_children

    if num_cpus == 0:
        # Non-bagged mode
        num_cpus = num_cpus_worker
        num_gpus = num_gpus_worker
    else:
        num_cpus_parent = num_cpus * num_parallel
        num_gpus_parent = num_gpus * num_parallel

        model_aux = model._user_params_aux
        if "num_cpus" not in model_aux:
            model_aux["num_cpus"] = num_cpus_parent
        if "num_gpus" not in model_aux:
            model_aux["num_gpus"] = num_gpus_parent

    upa = getattr(model, "model_base", model)._user_params_aux
    if "num_cpus" not in upa:
        getattr(model, "model_base", model)._user_params_aux["num_cpus"] = num_cpus
    if "num_gpus" not in upa:
        getattr(model, "model_base", model)._user_params_aux["num_gpus"] = num_gpus

    return model



import logging
import os
import time
from collections import OrderedDict
from copy import deepcopy
from typing import Tuple

from tqdm.auto import tqdm

from ..searcher import searcher_factory
from ..searcher.exceptions import ExhaustedSearchSpaceError
from ..searcher.local_searcher import LocalSearcher
from .reporter import FakeReporter

logger = logging.getLogger(__name__)


class LocalReporter:
    """
    Reporter implementation for LocalSequentialScheduler
    """

    def __init__(self, trial, searcher_config, training_history: dict, config_history: dict):
        self.trial = trial
        self.training_history = training_history
        self.training_history[trial] = []
        self.searcher_config = deepcopy(searcher_config)
        self.config_history = config_history
        self.trial_started = time.time()
        self.last_reported_time = self.trial_started
        self.last_result = None

    def __call__(self, *args, **kwargs):
        result = deepcopy(kwargs)
        if "done" not in result:
            result["trial"] = self.trial

            now = time.time()
            result["time_this_iter"] = now - self.last_reported_time
            result["time_since_start"] = now - self.trial_started
            self.last_reported_time = now

            self.training_history[self.trial].append(result)

            if self.trial not in self.config_history:
                self.config_history[self.trial] = self.searcher_config
                if "util_args" in self.searcher_config:
                    self.searcher_config.pop("util_args")

            self.last_result = result

    def terminate(self):
        pass  # compatibility


class LocalSequentialScheduler(object):
    """Simple scheduler which schedules all HPO jobs in sequence without any parallelism.
    The next trial scheduling will be decided based on the available time left within `time_out` setting
    and average time required for a trial to complete multiplied by the fill_factor (0.95) by default to
    accommodate variance in runtimes per HPO run.

    Parameters
    ----------
    train_fn : callable
        A task launch function for training.
    resource : dict
        Computation resources. For example, `{'num_cpus':2, 'num_gpus':1}`
    searcher : str
        Searcher (get_config decisions). If str, this is passed to
        searcher_factory along with search_options.
    search_options : dict
        If searcher is str, these arguments are passed to searcher_factory.
    num_trials : int
        Maximum number of jobs run in experiment. One of `num_trials`,
        `time_out` must be given.
    time_out : float
        If given, jobs are started only until this time_out (wall clock time).
        One of `num_trials`, `time_out` must be given.
    reward_attr : str
        Name of reward (i.e., metric to maximize) attribute in data obtained
        from reporter
    time_attr : str
        Name of resource (or time) attribute in data obtained from reporter.
        This attribute is optional for FIFO scheduling, but becomes mandatory
        in multi-fidelity scheduling (e.g., Hyperband).
        Note: The type of resource must be int.
    """

    def __init__(
        self,
        train_fn,
        search_space,
        train_fn_kwargs=None,
        searcher="auto",
        reward_attr="reward",
        resource=None,
        **kwargs,
    ):
        self.train_fn = train_fn
        self.training_history = None
        self.config_history = None
        self._reward_attr = reward_attr
        self.time_attr = kwargs.get("time_attr", None)
        self.resource = resource
        self.max_reward = kwargs.get("max_reward", None)
        self.searcher: LocalSearcher = self.get_searcher_(searcher, train_fn, search_space=search_space, **kwargs)
        self.init_limits_(kwargs)
        self.train_fn_kwargs = train_fn_kwargs
        self.metadata = {
            "search_space": search_space,
            "search_strategy": self.searcher,
            "stop_criterion": {"time_limit": self.time_out, "max_reward": self.max_reward},
            "resources_per_trial": self.resource,
        }

    def init_limits_(self, kwargs):
        if kwargs.get("num_trials", None) is None:
            assert kwargs.get("time_out", None) is not None, "Need stopping criterion: Either num_trials or time_out"
        self.num_trials = kwargs.get("num_trials", 9999)
        self.time_out = kwargs.get("time_out", None)
        if self.num_trials is None:
            assert self.time_out is not None, "Need stopping criterion: Either num_trials or time_out"

    def get_searcher_(self, searcher, train_fn, search_space, **kwargs) -> LocalSearcher:
        scheduler_opts = {}
        if searcher == "auto":
            searcher = "local_random"
            scheduler_opts = {"scheduler": "local"}
        elif searcher == "random":
            # FIXME: Hack to be compatible with gluoncv
            searcher = "local_random"

        search_options = kwargs.get("search_options", None)
        if isinstance(searcher, str):
            if search_options is None:
                search_options = dict()
            _search_options = search_options.copy()
            _search_options["search_space"] = search_space
            _search_options["reward_attribute"] = self._reward_attr
            # Adjoin scheduler info to search_options, if not already done by
            # subclass
            if "scheduler" not in _search_options:
                _search_options["scheduler"] = "local"
            searcher = searcher_factory(searcher, **{**scheduler_opts, **_search_options})
        else:
            assert isinstance(searcher, LocalSearcher)
        return searcher

    def run(self, **kwargs):
        """Run multiple trials given specific time and trial numbers limits."""
        self.searcher.configure_scheduler(self)

        self.training_history = OrderedDict()
        self.config_history = OrderedDict()

        failure_count = 0
        trial_count = 0
        trials_total_time = 0
        min_failure_threshold = 5
        failure_rate_threshold = 0.8
        time_start = time.time()

        r = range(self.num_trials)
        for i in tqdm(r) if self.num_trials < 1000 else r:
            trial_start_time = time.time()
            try:
                is_failed, result = self.run_trial(task_id=i)
            except ExhaustedSearchSpaceError:
                break
            except Exception:
                # TODO: Add special exception type when there are no more new configurations to try (exhausted search space)
                logger.log(30, f"\tWARNING: Encountered unexpected exception during trial {i}, stopping HPO early.")
                logger.exception("Detailed Traceback:")  # TODO: Avoid logging if verbosity=0
                break
            trial_end_time = time.time()

            trial_count += 1
            if is_failed:
                failure_count += 1
            else:
                trials_total_time += trial_end_time - trial_start_time

            if self.max_reward and self.get_best_reward() >= self.max_reward:
                logger.log(20, f"\tStopping HPO: Max reward reached")
                break

            if failure_count >= min_failure_threshold and (failure_count / trial_count) >= failure_rate_threshold:
                logger.warning(
                    f"Warning: Detected a large trial failure rate: "
                    f"{failure_count}/{trial_count} attempted trials failed ({round((failure_count / trial_count) * 100, 1)}%)! "
                    f"Stopping HPO early due to reaching failure threshold ({round(failure_rate_threshold * 100, 1)}%).\n"
                    f"\tFailures may be caused by invalid configurations within the provided search space."
                )
                break

            if self.time_out is not None:
                avg_trial_run_time = (
                    0 if trial_count == failure_count else trials_total_time / (trial_count - failure_count)
                )
                if not self.has_enough_time_for_trial_(
                    self.time_out, time_start, trial_start_time, trial_end_time, avg_trial_run_time
                ):
                    logger.log(20, f"\tStopping HPO to satisfy time limit...")
                    break

    @classmethod
    def has_enough_time_for_trial_(
        cls, time_out, time_start, trial_start_time, trial_end_time, avg_trial_run_time, fill_factor=0.95
    ):
        """
        Checks if the remaining time is enough to run another trial.

        Parameters
        ----------
        time_out total
            timeout in m
        time_start
            trials start time
        trial_start_time
            last trial start time
        trial_end_time
            last trial end time
        avg_trial_run_time
            running average of all trial runs
        fill_factor: float
            discount of `avg_trial_run_time` allowed for a next trial. Default is 0.95 of `avg_trial_run_time`

        Returns
        -------
            True if there is enough time to run another trial give runs statistics and remaining time

        """
        time_spent = trial_end_time - time_start
        is_timeout_exceeded = time_spent >= time_out
        time_left = time_start + time_out - trial_end_time
        is_enough_time_for_another_trial = True
        if avg_trial_run_time:
            is_enough_time_for_another_trial = time_left > avg_trial_run_time * fill_factor
        return is_enough_time_for_another_trial and not is_timeout_exceeded

    @classmethod
    def get_average_trial_time_(cls, i, avg_trial_run_time, trial_start_time, time_end):
        trial_time = time_end - trial_start_time
        if avg_trial_run_time is None:
            avg_trial_run_time = trial_time
        else:
            avg_trial_run_time = ((avg_trial_run_time * i) + trial_time) / (i + 1)
        return avg_trial_run_time

    def run_trial(self, task_id=0) -> Tuple[bool, dict]:
        """
        Start a trial with a given task_id

        Parameters
        ----------
        task_id
            task

        Returns
        -------
        is_failed: bool
            True if task completed successfully
        trial_start_time
            Trial start time
        trial_end_time
            Trial end time

        """
        new_searcher_config = self.searcher.get_config()
        searcher_config = deepcopy(self.metadata["search_space"])
        searcher_config.update(new_searcher_config)
        reporter = LocalReporter(task_id, searcher_config, self.training_history, self.config_history)
        return self.run_job_(task_id, searcher_config, reporter)

    def run_job_(self, task_id, searcher_config, reporter):
        args = dict()
        if self.train_fn_kwargs is not None:
            # TODO: Consider avoiding deepcopy and just do shallow copy,
            #  Risk is that it will allow values in self.train_fn_kwargs to be altered by HPO trials, causing early trials to alter later trials.
            train_fn_kwargs = deepcopy(self.train_fn_kwargs)
        else:
            train_fn_kwargs = dict()
        args.update(searcher_config)

        args["task_id"] = task_id
        self.searcher.register_pending(searcher_config)
        is_failed = False
        try:
            result = self.train_fn(args, reporter=reporter, **train_fn_kwargs)
            if type(reporter) is not FakeReporter:
                if reporter.last_result:
                    self.searcher.update(config=searcher_config, **reporter.last_result)
                else:
                    is_failed = True
        except Exception as e:
            logger.error(f"Exception during a trial: {e}")
            self.searcher.evaluation_failed(config=searcher_config)
            reporter(traceback=e)
            is_failed = True
            result = {"traceback": str(e)}
        return is_failed, result

    def run_with_config(self, config):
        """Run with config for final fit.
        It launches a single training trial under any fixed values of the hyperparameters.
        For example, after HPO has identified the best hyperparameter values based on a hold-out dataset,
        one can use this function to retrain a model with the same hyperparameters on all the available labeled data
        (including the hold out set). It can also returns other objects or states.
        """
        is_failed, result = self.run_job_("run_with_config", config, FakeReporter())
        return result

    def join_jobs(self, timeout=None):
        pass  # Compatibility

    def get_best_config(self):
        """Get the best configuration from the finished jobs."""
        # TODO: Consider passing the metadata search space to searcher to avoid having to do this
        searcher_config = deepcopy(self.metadata["search_space"])
        searcher_config.update(self.searcher.get_best_config())
        return searcher_config

    def get_best_reward(self):
        """Get the best reward from the finished jobs."""
        return self.searcher.get_best_reward()

    def get_training_curves(self, filename=None, plot=False, use_legend=True):
        """Get Training Curves"""
        if filename is None and not plot:
            logger.warning("Please either provide filename or allow plot in get_training_curves")
        import matplotlib.pyplot as plt

        eval_metric = self.__get_training_history_metric("eval_metric", default="validation_performance")
        sign_mult = int(self.__get_training_history_metric("greater_is_better", default=True)) * 2 - 1

        plt.ylabel(eval_metric)
        plt.xlabel(self.time_attr)
        plt.title("Performance vs Training-Time in each HPO Trial")
        for task_id, task_res in self.training_history.items():
            rewards = [x[self._reward_attr] * sign_mult for x in task_res]
            x = [x[self.time_attr] for x in task_res]
            plt.plot(x, rewards, label=f"task {task_id}")
        if use_legend:
            plt.legend(loc="best")
        if filename:
            logger.info(f"Saving Training Curve in {filename}")
            file_dir = os.path.split(os.path.abspath(filename))[0]
            if not os.path.exists(file_dir):
                os.makedirs(file_dir)
            plt.savefig(filename)
        if plot:
            plt.show()

    def __get_training_history_metric(self, metric, default=None):
        for _, task_res in self.training_history.items():
            if task_res and metric in task_res[0]:
                return task_res[0][metric]
        return default

    def get_best_task_id(self):
        """Get the task id that results in the best configuration/best reward.

        If there are duplicated configurations, we return the id of the first one.
        """
        best_config = self.get_best_config()
        for task_id, config in self.config_history.items():
            if best_config == config:
                return task_id
        raise RuntimeError(
            "The best config {} is not found in config history = {}. This should never happen!".format(
                best_config, self.config_history
            )
        )



class AutoGluonException(Exception):
    """
    Generic AutoGluon exception.
    Can be used to identify AutoGluon specific exception classes.
    """

    pass


class InsufficientTime(AutoGluonException):
    """
    Similar to TimeLimitExceeded, raised when the expected outcome of an operation
    would exceed the time limit, prior to exceeding the time limit.
    """

    pass


class TimeLimitExceeded(InsufficientTime):
    """
    Exception raised when the time limit has been exceeded (over budget)
    """

    pass


class NotEnoughMemoryError(AutoGluonException):
    pass


class NoGPUError(AutoGluonException):
    pass


class NotEnoughCudaMemoryError(AutoGluonException):
    pass


class NoValidFeatures(AutoGluonException):
    pass


class NoStackFeatures(NoValidFeatures):
    pass


class NotValidStacker(AutoGluonException):
    pass



from __future__ import annotations

import time
import typing

import pandas as pd

from ..models import AbstractModel
from ._abstract_callback import AbstractCallback

if typing.TYPE_CHECKING:
    # avoid circular import for type hints
    from ..trainer import AbstractTrainer


class ExampleCallback(AbstractCallback):
    """
    Example callback showcasing how to access and log information from the trainer.
    """

    def _before_model_fit(
        self,
        trainer: AbstractTrainer,
        model: AbstractModel,
        time_limit: float | None = None,
        stack_name: str = "core",
        level: int = 1,
        **kwargs,
    ) -> tuple[bool, bool]:
        time_limit_trainer = trainer._time_limit
        if time_limit_trainer is not None and trainer._time_train_start is not None:
            time_left_total = time_limit_trainer - (time.time() - trainer._time_train_start)
        else:
            time_left_total = None

        time_limit_log = f"\ttime_limit = {time_limit:.1f}\t(model)\n" if time_limit else ""
        time_limit_trainer_log = f"\ttime_limit = {time_limit_trainer:.1f}\t(trainer)\n" if time_limit_trainer else ""
        time_left_log = f"\ttime_left  = {time_left_total:.1f}\t(trainer)\n" if time_left_total else ""
        time_used_log = (
            f"\ttime_used  = {time_limit_trainer - time_left_total:.1f}\t(trainer)\n" if time_limit_trainer else ""
        )
        trainer.log(
            20,
            f"{self.__class__.__name__}.before_model_fit\n"
            f"\tmodel      = {model.name}\n"
            f"{time_limit_log}"
            f"{time_limit_trainer_log}"
            f"{time_left_log}"
            f"{time_used_log}"
            f"\tmodels_fit = {len(trainer.get_model_names())}\n"
            f"\tstack_name = {stack_name}\n"
            f"\tlevel      = {level}",
        )

        return False, False

    def _after_model_fit(
        self,
        trainer: AbstractTrainer,
        **kwargs,
    ) -> bool:
        with pd.option_context("display.max_rows", None, "display.max_columns", None, "display.width", 1000):
            trainer.log(20, f"{self.__class__.__name__}.after_model_fit | Leaderboard:\n{trainer.leaderboard()}")
        return False

    def before_trainer_fit(self, trainer: AbstractTrainer, **kwargs):
        trainer.log(20, f"{self.__class__.__name__}.before_trainer_fit | kwargs:\n")
        for k, v in kwargs.items():
            trainer.log(20, f"\t{k}={v}")

    def after_trainer_fit(self, trainer: AbstractTrainer):
        trainer.log(20, f"{self.__class__.__name__}.after_trainer_fit | Training Complete")
